/*
 * Command-line encoding utility for Ultima 4
 */

#include "lzw.h"
#include <stdlib.h>
#include <stdio.h>

long encodeU4File(char *outputFilename, char *deoutputFilename);
long getFilesize(FILE*);

int main (int argc, char *argv[])
{
    long errorCode;

    switch (argc)
    {
    case 3:
        errorCode = encodeU4File(argv[1], argv[2]);
        if (errorCode > 0)
        {
            return(EXIT_SUCCESS);
        }
        else
        {
            return(EXIT_FAILURE);
        }
        break;
    default:
        printf("Ultima 4 File Compression Utility\n");
        printf("Usage:\n");
        printf("arg1 = uncompressed input file\n");
        printf("arg2 = compressed output file\n");
        return(1);
        break;
    }
}

/*
 * loads a file, encodes it (from memory to memory), and writes the encoded data to another file
 * Returns:
 * -1 if there was an error
 * number of codewords in compressed file, on success
 */
long encodeU4File(char *inputFilename, char *outputFilename)
{
    long i;

    FILE *inputFile, *outputFile;
    unsigned char *inputMem, *outputMem;
    int *encodedMem;
    long inputFilesize, outputFilesize, codewordCount;
    long errorCode;

    if (strcmp (inputFilename, outputFilename) == 0) {
        printf("Input and output file names must not be identical.\n");
        return(-1);
    }
    else
    {
        if (!(inputFile=fopen(inputFilename,"rb"))||!(outputFile=fopen(outputFilename,"wb")))
        {
            printf("Couldn't open '%s' or '%s' or both.\n", inputFilename, outputFilename);
            return(-1);
        }
        else
        {
            /* size of the input file */
            inputFilesize = getFilesize(inputFile);

            if (inputFilesize == 0)
            {
                printf("Input file has a length of 0 bytes.\n");
                return(-1);
            }

            /* load input file into inputMem[] */
            inputMem = (unsigned char *) malloc(inputFilesize);
            fread(inputMem, 1, inputFilesize, inputFile);

            /* determine the number of 12-bit codewords in the compressed output */
            codewordCount = lzwGetCodewordCount(inputMem, inputFilesize);

            if (codewordCount > 0)
            {
                /* these variables are used by the packing code */
                long destination;
                long rootword;
                int shiftCount;

                encodedMem = (int *) malloc(sizeof(int[codewordCount]));

                /* compress inputMem[] to encodedMem[] */
                errorCode = lzwCompress(inputMem, encodedMem, inputFilesize);

                /* pack data from encodedMem[] to outputMem[] */
                outputFilesize = (codewordCount * 12 + 4) / 8;
                outputMem = (unsigned char *) malloc(outputFilesize);
                memset(outputMem, 0, outputFilesize);

                for (i = 0; i < codewordCount; i++)   /* pack the encoded data */
                {
                    destination = (i * 12) / 8;

                    shiftCount = (16-12)-((i*12)%8);
                    rootword = encodedMem[i] << shiftCount;
                    outputMem[destination + 0] |= rootword >> 8;
                    outputMem[destination + 1] |= rootword & 0xff;
                }

                /* write compressed output file */
                fwrite(outputMem, 1, outputFilesize, outputFile);

                /* clean up */
                fclose(outputFile);
                free(outputMem);
            }
            else
            {
                printf("input file has zero length.\n");

                /* delete the 0-byte file outputFile */
                fclose(outputFile);
                remove(outputFilename);

                return(-1);
            }
            
            /* clean up */
            free(inputMem);
            fclose(inputFile);
        }
    }

    return(errorCode);
}

/*
 * Returns the size of a file, and moves the file pointer to the beginning.
 * The file must already be open when this function is called.
 */
long getFilesize(FILE *input_file)
{
   long file_length;
   fseek(input_file, 0, SEEK_END);   /* move file pointer to file end */
   file_length = ftell(input_file);
   fseek(input_file, 0, SEEK_SET);   /* move file pointer to file start */
   return(file_length);
}
